// hacklu2013-fluxarchiv-bf-part1.c
//
// Brute-force the password of the archive for the first part of FluxArchiv
// exercise of hack.lu 2013 CTF.
//
// make hacklu2013-fluxarchiv-bf-part1 CFLAGS="--std=c99 -lcrypto -lpthread -Ofast -D_GNU_SOURCE"

#include <openssl/sha.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define CHARSET "ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789"
#define CHARSET_SZ (26 + 10)

#define TARGET "\x37\x29\x42\xdf\x27\x12\x82\x45\x05\xd8\x17\x1f\x4f\x0b\xcb\x14\x15\x3d\x39\xba"

#ifndef MULTITHREAD
# define MULTITHREAD 1
#endif

#if MULTITHREAD
# include <pthread.h>
# define WORKER_NB       8
# define WORKER_PASSWDS  50000000
#endif

#define PASSWD_SZ 6

#define MAX 0x14

int idx[MAX] = {0, 7, 14, 1, 8, 15, 2, 9, 16, 3, 10, 17, 4, 11, 18, 5, 12, 19, 6, 13};

void test_passwd(unsigned int tid, const char* passwd) {
    SHA_CTX sha1;
    char hash1[20], hash2[20], tmp[MAX];

    SHA1_Init(&sha1);
    SHA1_Update(&sha1, passwd, PASSWD_SZ);
    SHA1_Final(hash1, &sha1);

    for (int it = 0; it < MAX; ++it) {
        tmp[it] = hash1[idx[it]];
    }

    SHA1_Init(&sha1);
    SHA1_Update(&sha1, tmp, MAX);
    SHA1_Final(hash2, &sha1);

    if (memcmp(TARGET, hash2, MAX) == 0) {
        printf("[%02u] Found a matching password: %.6s\n", tid, passwd);
    }
}

int next_password(char passwd[PASSWD_SZ], int offsets[PASSWD_SZ]) {
    int it;

    for (it = 0; it < PASSWD_SZ && (it == 0 || offsets[it - 1] == 0); ++it) {
        offsets[it] += 1;                                                                                                                                                                         
        if (offsets[it] == CHARSET_SZ) {                                                                                                                                                          
            offsets[it] = 0;
        }                                                                                                                                                                                         
        passwd[it] = CHARSET[offsets[it]];
    }                                                                                                                                                                                             
    for (int it_ = it_; it_ < PASSWD_SZ; ++it_)
        passwd[it_] = CHARSET[offsets[it_]];
    return (it != PASSWD_SZ || offsets[PASSWD_SZ - 1] != 0);
}                                                                                                                                                                                                 

#if MULTITHREAD                                                                                                                                                                                   
struct thread_data {                                                                                                                                                                              
    unsigned int thread_id;
    unsigned int length;                                                                                                                                                                          
    char passwds[WORKER_PASSWDS][PASSWD_SZ];
};                                                                                                                                                                                                
                                                                                                                                                                                                  
int fetch_data(struct thread_data* data);

void* worker(void* _data) {
    struct thread_data* data = _data;
                                                                                                                                                                                                  
    if (data == NULL)                                                                                                                                                                             
        return NULL;

    while (fetch_data(data)) {
        //printf("[%02u] Thread is starting to process everything!\n", data->thread_id);
                                                                                                                                                                                                  
        for (unsigned int it = 0; it < data->length; ++it) {
            test_passwd(data->thread_id, data->passwds[it]);                                                                                                                                      
        }
        //printf("[%02u] Thread finished his batch!\n", data->thread_id);
    }                                                                                                                                                                                             
    //printf("[%02u] No job left for thread!\n", data->thread_id);
    return NULL;
}                                                                                                                                                                                                 
                                                                                                                                                                                                  
int fetch_data(struct thread_data* data) {
    static pthread_mutex_t mutex = PTHREAD_MUTEX_INITIALIZER;
    static int done = 0;                                                                                                                                                                          
    static int offsets[PASSWD_SZ] = {0};
    static unsigned long long int count = 0;

    int ret = 0;                                                                                                                                                                                  
                                                                                                                                                                                                  
    pthread_mutex_lock(&mutex);
    data->length = 0;
    if (done)                                                                                                                                                                                     
        goto end;                                                                                                                                                                                 
    for (int it = 0; it < WORKER_PASSWDS; ++it) {
        count += 1;
        if (next_password(data->passwds[it], offsets))
            data->length += 1;                                                                                                                                                                    
        else {
            //printf("[fetch_data] Count: %u\n", count);
            done = 1;                                            
            break;
        }
    }
    ret = data->length;
end:
    pthread_mutex_unlock(&mutex);
    return ret;
}
#endif

int main(void) {
#if MULTITHREAD
    pthread_t threadpool[WORKER_NB];
    struct thread_data* threaddata[WORKER_NB];

    /* Init thread data, should be done only once here. */
    for (int it = 0; it < WORKER_NB; ++it) {
        threaddata[it] = malloc(sizeof (struct thread_data));
        threaddata[it]->thread_id = it;
        for (int it_ = 0; it_ < WORKER_PASSWDS; ++it_) {
            memcpy(threaddata[it]->passwds[it_], "AAAAAA", 6);
        }
    }

    for (int it = 0; it < WORKER_NB; ++it) {
        pthread_create(&threadpool[it], NULL, worker, (void*) threaddata[it]);
    }

    for (int it = 0; it < WORKER_NB; ++it) {
        pthread_join(threadpool[it], NULL);
    }

    for (int it = 0; it < WORKER_NB; ++it)
        free(threaddata[it]);
#else
    char passwd[PASSWD_SZ + 1] = {0};
    static int offsets[PASSWD_SZ] = {0};

    while (next_password(passwd, offsets))
        test_passwd(0, passwd);
#endif
    return 0;
}
